; ---------------------------------------------------------------------------------------------------------
; HD44780-based LCD support for 6502 computer
  
; Display address
LCD_CMD  = $6F00
LCD_DATA = $6F01

  ; Instructions
  ; Clear display
LCD_CLR_DISPLAY = %00000001
  ; Return home
LCD_RET_HOME = %00000010
  ; Entry mode set
LCD_MODE_SET = %00000100
LCD_MODE_INCREMENT = %00000010
LCD_MODE_DECREMENT = %00000000
LCD_MODE_SHIFT = %00000001
  ; Display control
LCD_DISPLAY_CONTROL = %00001000
LCD_DISPLAY_ON = %00000100
LCD_DISPLAY_OFF = %00000000
LCD_DISPLAY_CURSOR_ON = %00000010  
LCD_DISPLAY_CURSOR_OFF = %00000000
LCD_DISPLAY_CURSOR_BLINK_ON = %00000001
LCD_DISPLAY_CURSOR_BLINK_OFF = %00000000
  ; Cursor or display shift
LCD_CURSOR_OR_DISPLAY_SHIFT = %00010000
LCD_CURSOR_MOVE = %00000000
LCD_CURSOR_DISPLAY_SHIFT = %00001000
LCD_CURSOR_LEFT = %00000000
LCD_CURSOR_RIGHT = %00000100
  ; Function set
LCD_FUNCTION_SET = %00100000
LCD_FUNCTION_4BITS = %00000000
LCD_FUNCTION_8BITS = %00010000
LCD_FUNCTION_2LINES = %00001000
LCD_FUNCTION_5BY8 = %00000000
  ; CGRAM (for character generation)
LCD_CGRAM_SET_ADDRESS = %01000000 ; Last 6 bits are the address
  ; DDRAM (for data)
LCD_DDRAM_SET_ADDRESS = %10000000 ; Last 7 bits are address
LCD_DDRAM_2ND_LINE_OFFSET = $40 ; This is the address of the first character on the second line

WAIT_CYCLES_4_1ms      = 30
WAIT_CYCLES_100us      = 1

lcd_print_debug:
  ; Prints string
  ; A=XX X=XX Y=XX to the LCD

  jsr lcd_print_accumulator

  pha
  phx
  phy
  phy
  phx

  ldx #<x_eq
  ldy #>x_eq
  jsr lcd_print_string

  plx
  txa
  jsr to_hex
  jsr lcd_print_string

  ldx #<y_eq
  ldy #>y_eq
  jsr lcd_print_string

  ply
  tya
  jsr to_hex
  jsr lcd_print_string

  ply
  plx
  pla
  rts

lcd_print_accumulator:
  ; Prints content of the accumulator as hexadecimal
  ; number at current position on the LCD (' A=xx')

  phx
  phy

  ldx #<a_eq
  ldy #>a_eq
  jsr lcd_print_string

  jsr to_hex
  jsr lcd_print_string

  ply
  plx
  rts

a_eq  .asciiz " A="
x_eq  .asciiz " X="
y_eq  .asciiz " Y="


lcd_print_string:
  ; Prints a null terminated string
  ; to the LCD at current position
  ;
  ; PARAMETERS
  ; X, Y - Address of the string (X - Low, Y - High byte)

  stx LCD_STRING_ADDRESS
  sty LCD_STRING_ADDRESS + 1

  pha
  ldy #0

.loop:
  lda (LCD_STRING_ADDRESS), y
  beq .end

  jsr lcd_write_data
  jsr lcd_wait_busy

  iny
  jmp .loop

.end:
  pla
  rts


lcd_goto:
  ; Change line and column.
  ;
  ; PARAMETERS
  ; X - Column
  ; Y - Row

  pha

  txa                 ; Copy column to A

  cpy #0              ; Compare row to 0
  beq .write_command  ; If 0 then just write command
  
  ; Row is one
  clc
  adc #LCD_DDRAM_2ND_LINE_OFFSET

.write_command:
  ora #LCD_DDRAM_SET_ADDRESS
  jsr lcd_write_command

  pla
  rts


lcd_home:
  ; Return home
  ; Set cursor at 0,0

  pha

  lda #LCD_RET_HOME
  jsr lcd_write_command
  jsr lcd_wait_busy

  pla
  rts


lcd_clear:
  ; Clear the display

  pha

  lda #LCD_CLR_DISPLAY
  jsr lcd_write_command
  jsr lcd_wait_busy

  pla
  rts

lcd_clear_line:
  ; Clears given line
  ;
  ; PARAMETERS
  ; Y - line to be cleared

  phx
  phy

  ldx #0
  jsr lcd_goto

  ldx #<.clr_string
  ldy #>.clr_string
  jsr lcd_print_string

  ldx #0
  ply
  jsr lcd_goto

  plx
  rts

.clr_string .asciiz "                "

lcd_init:
  ; Initialize the LCD

  pha

  lda #(LCD_FUNCTION_SET | LCD_FUNCTION_8BITS)
  sta LCD_CMD
  ; Wait 4.1 ms
  lda #WAIT_CYCLES_4_1ms
  jsr delay_a

  lda #(LCD_FUNCTION_SET | LCD_FUNCTION_8BITS)
  sta LCD_CMD
  ; Wait 100 us
  lda #WAIT_CYCLES_100us
  jsr delay_a

  ; BF can be checked after this instruction
  lda #(LCD_FUNCTION_SET | LCD_FUNCTION_8BITS)
  sta LCD_CMD
  lda #WAIT_CYCLES_100us
  jsr delay_a
  ; jsr lcd_wait_busy

  ; Function set (4-bits). Interface is still 8-bits long.
  lda #(LCD_FUNCTION_SET | LCD_FUNCTION_4BITS)
  sta LCD_CMD
  jsr lcd_wait_busy

  ; Function set (8-bits, 2 lines, 5x8 chars)
  lda #(LCD_FUNCTION_SET | LCD_FUNCTION_4BITS | LCD_FUNCTION_2LINES | LCD_FUNCTION_5BY8)
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Display off
  lda #(LCD_DISPLAY_CONTROL)
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Display clear
  lda #LCD_CLR_DISPLAY
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Mode set (incrementing)
  lda #(LCD_MODE_SET | LCD_MODE_INCREMENT)
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Cursor
  lda #(LCD_CURSOR_OR_DISPLAY_SHIFT | LCD_CURSOR_MOVE | LCD_CURSOR_RIGHT)
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Display on w/ blinking cursor
  lda #(LCD_DISPLAY_CONTROL | LCD_DISPLAY_ON | LCD_DISPLAY_CURSOR_ON | LCD_DISPLAY_CURSOR_BLINK_ON)
  ; lda #(LCD_DISPLAY_CONTROL | LCD_DISPLAY_ON)
  jsr lcd_write_command
  jsr lcd_wait_busy

  ; Return home
  lda #LCD_RET_HOME
  jsr lcd_write_command
  jsr lcd_wait_busy

  pla
  rts


lcd_wait_busy:
  ; Wait for busy flag to clear

  pha

.loop:
  jsr lcd_read_command
  and #%10000000
  bne .loop

  pla
  rts


lcd_write_command:
  ; Sends current content of the accumulator
  ; to the LCD using 4 bit mode (command port)
  ;
  ; PARAMETERS
  ; A  - value to be sent
  pha

  sta LCD_CMD ; Send high nibble
  asl
  asl
  asl
  asl         ; Shift 4 bits left (low nibble is high now)
  sta LCD_CMD ; Send low nibble

  pla
  rts


lcd_write_data:
  ; Same as above but with data port
  
  pha

  sta LCD_DATA ; Send high nibble
  asl
  asl
  asl
  asl          ; Shift 4 bits left (low nibble is high now)
  sta LCD_DATA ; Send low nibble

  pla
  rts


lcd_read_command:
  ; Reads content into the accumulator
  ; from the LCD using 4 bit mode (command port)
  ;
  ; RETURNS
  ; A - value read

  lda LCD_CMD    ; Load high nibble
  sta LCD_NIBBLE ; Store high nibble in RAM

  lda LCD_CMD    ; Load low nibble into high nibble
  lsr
  lsr
  lsr
  lsr            ; Low nibble is now low

  ora LCD_NIBBLE   ; OR low nibble with high nibble in RAM

  rts


lcd_read_data:
  ; Same as above but with data port

  lda LCD_DATA   ; Load high nibble
  sta LCD_NIBBLE ; Store high nibble in RAM

  lda LCD_DATA   ; Load low nibble into high nibble
  lsr
  lsr
  lsr
  lsr            ; Low nibble is now low

  ora LCD_NIBBLE ; OR low nibble with high nibble in RAM

  rts