  .include "../symbols_app.s"

; Begin ROM
  .org $8000

; Indirect vectors
APP_INIT_BYTES = $3000    ; 2 bytes - Init bytes, first 2 bytes should be set to AA 55 for the app to start.
APP_SPACE_VECTOR = $3002  ; 2 bytes
IRQ_VECTOR = $3004        ; 2 bytes
NMI_VECTOR = $3006        ; 2 bytes

reset:
  ; Reset vector
  ; Initialize stack pointer
  ldx #$FF
  txs

  ; Initialize LCD
  jsr lcd_init

  ldx #<init_string
  ldy #>init_string
  jsr lcd_print_string

  ldx #0
  ldy #1
  jsr lcd_goto

  ; Initialize UART
  jsr uart_init

loader_init:
  ; 16 tries x 60 ms = 0.96 second for receiving valid init signal
  ldx #16

program_loop:
  lda #'+'                        ; Send init sign
  jsr uart_send

  ; Receive command id
  ldy #6                          ; 6 tries - 60 ms
  jsr uart_recv_with_timeout      ; Receive with timeout 100 ms

  cpy #0
  beq .success

  ; If Y not 0 - failed
  lda #'-'
  jsr lcd_write_data              ; Print dash at second line

  dex
  cpx #0                          ; If counter not zero
  bne program_loop               ; Goto program loop

  ; End loop - failed to receive - timeout
  ldx APP_INIT_BYTES
  ldy APP_INIT_BYTES+1            ; Load init bytes

  cpx #$AA                        ; Compare first byte to AA
  bne .no_app
  cpy #$55                        ; Compare second byte to 55
  bne .no_app

  ; Two bytes exact match
  jsr lcd_clear                   ; Clear LCD
  jmp (APP_SPACE_VECTOR)          ; Jump to app

.no_app:
  ldx #5
  ldy #1
  jsr lcd_goto

  ldx #<no_app_string
  ldy #>no_app_string
  jsr lcd_print_string            ; Print no app after the dashes

  ldx #28
  lda #155
  jsr delay_x                     ; Delay 1s

  ldy #1
  jsr lcd_clear_line              ; Clear second line

  jmp loader_init

.success:
  ; Receive valid init signal

  jsr lcd_write_data              ; Print received instruction

  cmp #"w"                        ; Write page command
  beq write_page

  cmp #"r"                        ; Read page command
  beq read_page

  ; No match
  ; Print error message
  lda #"e"
  jsr uart_send
  
  jmp loader_init


write_page:
  ; Writes a page
  ; Needs to be sent a page address (1 byte)
  ; and 256-byte page data

  pha
  phy

  ; Receive page address
  jsr uart_recv

  ; Setup page write address
  stz BTLDR_PAGE_ADDRESS        ; Store low byte first (always zero)
  sta BTLDR_PAGE_ADDRESS + 1    ; Store high byte
  ldy #0                        ; Clear Y index

.loop:
  jsr uart_recv
  sta (BTLDR_PAGE_ADDRESS), y

  iny
  bne .loop
  ; End loop

  ply
  pla
  jmp program_loop


read_page:
  ; Reads a page
  ; Needs to be sent a page address (1 byte)
  ; 256-byte page data will be sent

  pha
  phy

  ; Receive page address
  jsr uart_recv

  ; Setup page read address
  stz BTLDR_PAGE_ADDRESS        ; Store low byte first (always zero)
  sta BTLDR_PAGE_ADDRESS + 1    ; Store high byte 
  ldy #0                        ; Clear Y index

.loop:
  lda (BTLDR_PAGE_ADDRESS), y
  jsr uart_send
  
  iny
  bne .loop

  ply
  pla

  dex
  jmp program_loop


; Strings
init_string .asciiz "6502-btldr init"
no_app_string .asciiz "no app"

; Include files
  .include "../variables.s"

  .include "delay.s"
  .include "number.s"

  .include "uart.s"
  .include "HD44780.s"

  .include "lib/ps2.s"
  .include "lib/vga.s"
  .include "lib/string.s"
  .include "lib/printf.s"
  .include "lib/pstack.s"
  .include "lib/math/hex.s"
  .include "lib/math/modulo.s"
  .include "lib/math/decimal.s"
  .include "lib/math/smallfloat.s"

; Interrupt routines
irq:
  pha

  lda PS2_STATUS_REGISTER     ; Load PS/2 status
  sta PS2_STATUS              ; Save PS/2 status
  and #PS2_INT_MASK           ; Check PS/2 interrupt source
  bne .ps2_interrupt
  jmp .end

.ps2_interrupt:
  jsr ps2_interrupt_handler
  jmp .end

.end:
  pla
  jmp (IRQ_VECTOR)

nmi:
  jmp (NMI_VECTOR)

; Interrupt vectors
  .org $FFFA
  .word nmi
  .word reset
  .word irq